package charlesgunn.anim.core;

import java.awt.Color;
import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.List;
import java.util.Vector;
import java.util.WeakHashMap;

import charlesgunn.anim.io.ImportExport;
import de.jreality.scene.Camera;

/**
 * This class uses Java bean introspection to animate the visible properties of 
 * an instance of an arbitrary Java class.  Only those properties of known type will be
 * animated.  Currently that includes Double, java.awt.Color, and de.jreality.scene.Transformation.
 * 
 * Properties that shouldn't be animated can be excluded by using the constructor 
 * {@link #KeyFrameAnimatedBean(Object, List)}.
 * 
 * For each non-excluded property of known type, a key frame animated variable of the 
 * appropriate type is created.  Method calls to methods in the {@link KeyFrameAnimated} interface are
 * then propagated to this list of animated variables.  Subclasses that need to do something
 * additional in these methods need to subclass these methods.
 * 
 * TODO: provide control of interpolation type and boundary mode
 * @author Charles Gunn
 *
 * @param <T>  The class of the underlying object being animated.
 */
public class KeyFrameAnimatedBean<T> implements KeyFrameAnimated<T> {
	transient BeanInfo info;
	static int count = 0;
	String name = "unnamedBean"+count++;
	protected List<String> excluded = new Vector<String>();
	protected List<KeyFrameAnimated> animatedFields = new Vector<KeyFrameAnimated>();
	transient WeakHashMap<PropertyDescriptor, KeyFrameAnimated<?>> prop2Anim = new WeakHashMap<PropertyDescriptor, KeyFrameAnimated<?>>();
	transient T target;
	transient Runnable updateRunnable = null;
	boolean onlyAddChanges = false,
			givesWay = true;
	
	public KeyFrameAnimatedBean(T t, List<String> exc)	{
	   	target = t;
	   	if (target == null) return;
	   	if (exc != null) excluded = exc;
		getBeanInfo();
		setupAnimation();		
	}
    public KeyFrameAnimatedBean(T t) {
    	this(t, null);
 	}

	private void setupAnimation() {
		for ( PropertyDescriptor pd : info.getPropertyDescriptors() ) {
			if (excluded.contains(pd.getName())) continue;
        	Class<?> c = pd.getPropertyType();
//        	System.err.println("Property is "+pd.getName());
           	final Method rm = pd.getReadMethod();
        	final Method wm = pd.getWriteMethod();
        	if (rm == null || wm == null) continue;
        	if (c.equals(Double.class) || c.equals(double.class))	{
        		KeyFrameAnimatedDouble foo = new KeyFrameAnimatedDouble(new AnimatedBeanDelegate<Double>(rm, wm));
        		foo.setName(pd.getName());
        		animatedFields.add(foo);
        		prop2Anim.put(pd, foo);
//        		System.err.println("Adding animated double for field "+pd.getName());
        	} 
        	if (c.equals(Integer.class) || c.equals(int.class))	{
        		KeyFrameAnimatedInteger foo = new KeyFrameAnimatedInteger(new AnimatedBeanDelegate<Integer>(rm, wm));
        		foo.setName(pd.getName());
        		animatedFields.add(foo);
        		prop2Anim.put(pd, foo);
//        		System.err.println("Adding animated integer for field "+pd.getName());
        	} 
        	else if (c.equals(Color.class) )	{
        		KeyFrameAnimatedColor foo = new KeyFrameAnimatedColor(new AnimatedBeanDelegate<Color>(rm, wm));
        		foo.setName(pd.getName());
        		animatedFields.add(foo);
        		prop2Anim.put(pd, foo);
//       		System.err.println("Adding animated Color for field "+pd.getName());
        	} 
        	else if (c.equals(Boolean.class) || c.equals(boolean.class))	{
        		KeyFrameAnimatedBoolean foo = new KeyFrameAnimatedBoolean(new AnimatedBeanDelegate<Boolean>(rm, wm));
        		foo.setName(pd.getName());
        		animatedFields.add(foo);
        		prop2Anim.put(pd, foo);
//       		System.err.println("Adding animated boolean for field "+pd.getName());
        	} 
//        	System.err.println("Adding animated "+c+" for field "+pd.getName());
      }
	}

	BeanInfo getBeanInfo() {
		if (info != null) return info;
		try {
			info = Introspector.getBeanInfo( target.getClass() );
		} catch (IntrospectionException e) {
			e.printStackTrace();
		}
		return info;
	}

	public String getName() {
		if (name == null) name = "AnimatedBean "+count++;
		return name;
	}

	public void setName(String n) {
		name = n;
	}

	public void setCurrentValue( T value) {
		for (PropertyDescriptor pd : prop2Anim.keySet())	{
			System.err.println("setCurrentValue "+pd.getName());
			KeyFrameAnimated an = prop2Anim.get(pd);
	       	Method rm = pd.getReadMethod();
        	Method wm = pd.getWriteMethod();
        	if (rm == null || wm == null) continue;
        	try {
				an.setCurrentValue(rm.invoke(value, null));
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				e.printStackTrace();
			}
			
		}
	}

	public void addKeyFrame(TimeDescriptor td) {
		if (!writable) return;
		for (KeyFrameAnimated<?> an : animatedFields)	
			an.addKeyFrame(td);
	}

	public void addKeyFrame(KeyFrame<T> kf) {
		if (!writable) return;
		setCurrentValue(kf.getValue());
		addKeyFrame(kf.getTimeDescriptor());
	}
	
	public void deleteKeyFrame(TimeDescriptor time) {
		for (KeyFrameAnimated<?> an : animatedFields)	
			an.deleteKeyFrame(time);
	}

	public void setValueAtTime(double t) {
		for (KeyFrameAnimated<?> an : animatedFields)	
			an.setValueAtTime(t);	
		if (updateRunnable != null) updateRunnable.run();
	}
	
	public void setFromArchive(KeyFrameAnimated<T> archived) {
		if ( !(archived instanceof KeyFrameAnimatedBean))
			throw new IllegalArgumentException("Wrong class");
		KeyFrameAnimatedBean<T> archBean = (KeyFrameAnimatedBean<T>) archived;
		excluded = archBean.excluded;
		for (KeyFrameAnimated foo : archBean.animatedFields) {
			if ( !(foo instanceof SimpleKeyFrameAnimated)) {
				System.err.println("Can't handle class "+foo.getClass());
				continue;
			}
			SimpleKeyFrameAnimated akfa = (SimpleKeyFrameAnimated) foo;
			SimpleKeyFrameAnimated thisAkfa = (SimpleKeyFrameAnimated) 
				ImportExport.elementNamed(akfa.getName(), animatedFields);
			if (thisAkfa == null)	{
				System.err.println("Can't find animated named "+akfa.getName());
				continue;
			}
//			System.err.println("Handling bean property "+akfa.getName());
			thisAkfa.setFromArchive(akfa);
		}
	}

	/**
	 * An instance of this class handles one property of the bean, by using the
	 * read and write method to implement the gather and propagate methods.
	 * 
	 * @author Charles Gunn
	 *
	 * @param <X>	The type of the property being animated
	 */
	class AnimatedBeanDelegate<X> implements KeyFrameAnimatedDelegate<X> {
		Method rm, wm;
		
		AnimatedBeanDelegate(Method r, Method w) {
			rm = r;
			wm = w;
		}
		
		public X gatherCurrentValue(X t) {
			X d = null;
			try {
				d = (X) rm.invoke(target, null);
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				e.printStackTrace();
			}	
			return d;
		}

		public void propagateCurrentValue(X t) {
			try {
				wm.invoke(target, t);
			} catch (IllegalArgumentException e) {
				e.printStackTrace();
			} catch (IllegalAccessException e) {
				e.printStackTrace();
			} catch (InvocationTargetException e) {
				e.printStackTrace();
			}
		}
		
	}

	public static void main( String[] args )
	   throws IntrospectionException
	   {
		   Camera c = new Camera();
		   c.setNear(3);
		   c.setFar(5.0);
		   c.setPerspective(false);
		   c.setName("harvey");
		   KeyFrameAnimatedBean<Camera> ac = new KeyFrameAnimatedBean<Camera>(c) {	   
			   @Override
			   public void setValueAtTime( double t)	{
				   super.setValueAtTime(t);
//				   System.err.println("Updating");
			   }
	
		   };
		   System.err.println("Copy "+c.getName()+"Near = "+c.getNear());
	}
	public Runnable getUpdateRunnable() {
		return updateRunnable;
	}
	public void setUpdateRunnable(Runnable updateRunnable) {
		this.updateRunnable = updateRunnable;
	}
	public List<String> getExcluded() {
		return excluded;
	}
	public void setExcluded(List<String> excluded) {
		this.excluded = excluded;
	}
	public List<KeyFrameAnimated> getAnimatedFields() {
		return animatedFields;
	}
	public void setAnimatedFields(List<KeyFrameAnimated> animatedFields) {
		this.animatedFields = animatedFields;
	}
	boolean writable = true;
	public boolean isWritable() {
		return writable;
	}
	public void setWritable(boolean b) {
		writable = b;
	}
	public void printState() {
		// TODO Auto-generated method stub
		
	}
	public void endAnimation() {
		// TODO Auto-generated method stub
		
	}
	public void startAnimation() {
		// TODO Auto-generated method stub
		
	}
	public boolean isOnlyAddChanges() {
		return onlyAddChanges;
	}
	public void setOnlyAddChanges(boolean b) {
		onlyAddChanges = b;
	}
	public boolean isEqual(T v1, T v2) {
		return false;
	}
	@Override
	public boolean isGivesWay() {
		// TODO Auto-generated method stub
		return false;
	}
	@Override
	public void setGivesWay(boolean b) {
		givesWay = b;
		for (KeyFrameAnimated<?> an : animatedFields)	
			an.setGivesWay(b);	

	}


}
