package charlesgunn.anim.core;

import charlesgunn.anim.util.AnimationUtility;
import charlesgunn.anim.util.AnimationUtility.InterpolationTypes;
import de.jreality.math.CubicBSpline;
import de.jreality.math.FactoredMatrix;
import de.jreality.math.Quaternion;
import de.jreality.math.Rn;

/**
 * The basic class for interpolating isometries in euclidean and non-euclidean spaces.
 * Uses an instance of {@link FactoredMatrix} to manage the factoring and interpolation of
 * values.  Specific classes that represent such isometries should subclass this one by
 * providing a delegate.  See {@link KeyFrameAnimatedTransformation} for an example.
 * @author Charles Gunn
 *
 */
public class KeyFrameAnimatedIsometry extends SimpleKeyFrameAnimated<FactoredMatrix> {
	
	public KeyFrameAnimatedIsometry()	{
		this(new FactoredMatrix());
	}
	
	public KeyFrameAnimatedIsometry(FactoredMatrix t) {
		super(t);
	}
	
	public KeyFrameAnimatedIsometry(KeyFrameAnimatedDelegate<FactoredMatrix> animatedDelegate) {
		super(animatedDelegate);
	}
	
	@Override
	public FactoredMatrix copyFromTo(FactoredMatrix from, FactoredMatrix to) {
		to.assignFrom(from);
		to.update();
		return to;
	}

	@Override
	public FactoredMatrix getNewInstance() {
		if (target != null) return new FactoredMatrix(target.getMetric());
		return new FactoredMatrix();
	}

	@Override
	public void setValueAtTime(double t) {
		if (getKeyFrames().size() == 0) return;
   		if (getKeyFrames().size() > 1 && 
   			interp == InterpolationTypes.CUBIC_BSPLINE &&
   			keyFramesChanged())	{
	   				updateBSplines();
	   				keyFramesChanged = false;
   			}
		super.setValueAtTime(t);
		t = remappedTime;
		if (!valueIsSet)  
			if (interp == InterpolationTypes.CUBIC_BSPLINE && getKeyFrames().size() > 1)	{
		      	currentValue.setTranslation(cbx.valueAt(t), cby.valueAt(t), cbz.valueAt(t));	
		      	Quaternion rot = new Quaternion();
				rot.re = cbr.valueAt(t);
				rot.x = cbi.valueAt(t);
				rot.y = cbj.valueAt(t);
				rot.z = cbk.valueAt(t);
				Quaternion.normalize(rot, rot);
		      	currentValue.setRotation(rot);	
		      	currentValue.setStretch(cbsx.valueAt(t), cbsy.valueAt(t), cbsz.valueAt(t));
			} else if (interp == InterpolationTypes.CUBIC_HERMITE && getKeyFrames().size() > 1) {
				double hermiteT = AnimationUtility.hermiteInterpolation(t, previous.getTime(), next.getTime(), previous.getTime(),next.getTime());	
//				System.err.println("time = "+t+"hermite time = "+hermiteT);
				currentValue = AnimationUtility.linearInterpolation(
						currentValue, hermiteT, previous.getTime(),next.getTime(), previous.getValue(), next.getValue());
			} else
				currentValue = AnimationUtility.linearInterpolation(
						currentValue, t, previous.getTime(),next.getTime(), previous.getValue(), next.getValue());
		if (target != null) {
			target.assignFrom(currentValue);
			target.update();
		}
		delegate.propagateCurrentValue(currentValue);
	}
	
	transient CubicBSpline cbx, cby, cbz, cbr, cbi, cbj, cbk, cbsx, cbsy, cbsz;
	boolean useNatural = false;
	void updateBSplines()	{
		lastModified = System.currentTimeMillis();
		if (getKeyFrames().size() <= 1) return;
		int n = getKeyFrames().size();
		double[] t = new double[n];
		double[] x = new double[n];
		double[] y = new double[n];
		double[] z = new double[n];
		double[] r = new double[n];
		double[] qi = new double[n];
		double[] qj = new double[n];
		double[] qk = new double[n];
		double[] sx = new double[n];
		double[] sy = new double[n];
		double[] sz = new double[n];
		Quaternion oldQ = null;
  		for (int i = 0; i<getKeyFrames().size(); ++i)	{
			t[i] = getKeyFrames().get(i).getTime();
  			FactoredMatrix fm = getKeyFrames().get(i).getValue();
 			double[] m = fm.getArray();
			Quaternion rot = fm.getRotationQuaternion();
			double w = m[15];
			x[i] = m[3]/w;
			y[i] = m[7]/w;
			z[i] = m[11]/w;
			// try to stay on the same sheet
			if (oldQ != null) {
				double dot = Quaternion.innerProduct(oldQ, rot);
				dot = Math.max(-1.0, Math.min(1.0, dot));
//				System.err.println("angle is "+Math.acos(Math.abs(dot)));
				if (dot < 0) Quaternion.times(rot, -1.0, rot);
			}
			r[i] = rot.re;
			qi[i] = rot.x;
			qj[i] = rot.y;
			qk[i] = rot.z;
			oldQ = rot;
			double[] scale = fm.getStretch();
			sx[i] = scale[0];
			sy[i] = scale[1];
			sz[i] = scale[2];
		}
   		if (useNatural)	{
			cbx = new CubicBSpline.Natural(t, x);
			cby = new CubicBSpline.Natural(t, y);
			cbz = new CubicBSpline.Natural(t, z);
			cbr = new CubicBSpline.Natural(t, r);
			cbi = new CubicBSpline.Natural(t, qi);
			cbj = new CubicBSpline.Natural(t, qj);
			cbk = new CubicBSpline.Natural(t, qk);
    		cbsx = new CubicBSpline.Natural(t, sx);
    		cbsy = new CubicBSpline.Natural(t, sy);
    		cbsz = new CubicBSpline.Natural(t, sz);        				
   		} else {
   			cbx = new CubicBSpline.Default(t, x);
			cby = new CubicBSpline.Default(t, y);
			cbz = new CubicBSpline.Default(t, z);
			cbr = new CubicBSpline.Default(t, r);
			cbi = new CubicBSpline.Default(t, qi);
			cbj = new CubicBSpline.Default(t, qj);
			cbk = new CubicBSpline.Default(t, qk);  
    		cbsx = new CubicBSpline.Default(t, sx);
    		cbsy = new CubicBSpline.Default(t, sy);
    		cbsz = new CubicBSpline.Default(t, sz);        				
   		}



	}

	private Object readResolve()	{
		updateBSplines();
		return this;
	}

	@Override
	public boolean equalValues(FactoredMatrix t1, FactoredMatrix t2) {
		double[] array = t1.getArray();
		double[] array2 = t2.getArray();
		System.err.println("t1 array = "+Rn.matrixToJavaString(array));
		System.err.println("t2 array = "+Rn.matrixToJavaString(array2));
		return Rn.equals(array, array2, 10E-8);
	}

}
