package charlesgunn.anim.core;

import java.util.List;

import charlesgunn.anim.util.AnimationUtility.BoundaryModes;
import charlesgunn.anim.util.AnimationUtility.InterpolationTypes;
import de.jreality.util.LoggingSystem;


/**
 * This class implements {@link KeyFrameAnimated}.  It handles generic bookkeeping
 * like managing a sorted list of key frames. It comes with two constructors corresponding
 * to the two main usage patterns: for a mutable type, give an instance of the type
 * which will then be used for reading (setting key frames) and writing (playback).
 * For immutable types, there is a constructor requiring an instance of 
 * {@link KeyFrameAnimatedDelegate} which takes care of this reading and writing.
 * (All instances have delegates, but in the first case the delegate is created automatically.)
 * 
 * In order to be able to construct key-value pairs, this class requires that
 * subclasses implements a copy method {@link #copyFromTo(Object, Object)}) and provide fresh
 * instances {@link #getNewInstance()}.
 * 
 * TODO: make a clearer distinction between immutable and mutable classes.  Possibly make two
 * classes.
 * 
 * @author Charles Gunn
 *
 * @param <T>	T is the class of the underlying class which is being animated
 */

abstract public class SimpleKeyFrameAnimated<T> implements KeyFrameAnimated<T> {
	
	static int count = 0;
	protected String name = "keyFrameAnimated"+count++;
	protected SortedKeyFrameList<T> keyFrames = new SortedKeyFrameList<T>();
	protected InterpolationTypes interp = InterpolationTypes.LINEAR;
	protected BoundaryModes boundaryMode = BoundaryModes.CLAMP;
	protected boolean onlyAddChanges = true,
			givesWay = false;
	protected transient TimeMapper timeMapper = null;
	protected transient double remappedTime;
	protected transient T target, currentValue;
	protected transient KeyFrameAnimatedDelegate<T> delegate;
	protected transient boolean keyFramesChanged = true;
	protected boolean writable = true;
	protected transient boolean debug = false;
	transient long lastModified = 0;
	
	public SimpleKeyFrameAnimated(KeyFrameAnimatedDelegate<T> ad)	{
		setDelegate(ad);
		target = currentValue = getNewInstance();
	}
	
	/**
	 * Behind the scenes, all instances have an animated delegate, but
	 * in this case we can generate it automatically.
	 * @param t
	 */
	public SimpleKeyFrameAnimated(T t)	{
		target = t;
		currentValue = copyFromTo(t, getNewInstance());
		setDelegate(getDefaultDelegate());
	}
	
	public SimpleKeyFrameAnimated() {
		currentValue = getNewInstance();
	}

	abstract public T copyFromTo(T from, T to);
	
	abstract protected T getNewInstance(); 
	
	abstract public boolean equalValues(T t1, T t2);
	
	public KeyFrameAnimatedDelegate<T> getDelegate() {
		return delegate;
	}

	public void setDelegate(KeyFrameAnimatedDelegate<T> delegate) {
		this.delegate = delegate;
	}
	
	/**
	 * In the case of mutable objects, we can provide a standard delegate
	 * as long was we were given a target instance in the constructor.
	 * @return
	 */
	public KeyFrameAnimatedDelegate<T> getDefaultDelegate()	{
		if (target == null) return null;
		return new KeyFrameAnimatedDelegate<T>() {

			public T gatherCurrentValue(T t) {
				return copyFromTo(target, t);
			}

			public void propagateCurrentValue(T t) {
				copyFromTo(t, target);
			}
			
		};
	}

	public void deleteKeyFrame(TimeDescriptor td) {
		// more likely that the time descriptor is the same
		KeyFrame<T> foo = keyFrames.existingKeyFrameAt(td);
		if (foo != null) {
			keyFrames.remove(foo);
			keyFramesChanged = true;
//			System.err.println(name+" deleting key frame at "+td.getTime());
		} else 
			//throw new IllegalStateException("No such key frame");
			LoggingSystem.getLogger(this).warning("No such key frame "+td.getTime());
	}

	/**
	 * TODO figure out how to make this method work for both 
	 * immutable and mutable classes T.  Currently it only works with
	 * immmutable classes, for mutable classes the target itself should
	 * be set.  Can I just set the value of the target in this case?
	 */
	public void setCurrentValue( T value) {
		currentValue = copyFromTo(value, getNewInstance());
	}

	public void addKeyFrame(TimeDescriptor td) {
		if (!writable) return;
		T copy  = copyFromTo(currentValue, getNewInstance());
		copy = delegate.gatherCurrentValue(copy);

		addKeyFrame(new KeyFrame<T>(td, copy));
	}

	public void addKeyFrame(KeyFrame<T> newKF) {
		if (!writable) return;
		KeyFrame<T> foo = keyFrames.existingKeyFrameAt(newKF.getTimeDescriptor());
		if (foo != null)  {
			if (givesWay)	{
				System.err.println("giving way "+getName());
				KeyFrame<T> iter = foo;
				T oldValue = copyFromTo(foo.getValue(), getNewInstance());
				// if the new value is the same as the old value do nothing
				if (equalValues(oldValue, newKF.getValue())) return;
				int count = keyFrames.indexOf(foo);
				// otherwise go through list and replace oldvalue with the new one
				// until something different appears
				do {
					System.err.println("giving way "+count);
					iter.setValue(newKF.getValue());
					count++;
					if (count == keyFrames.size()) break;
					iter = keyFrames.get(count);
					System.err.println("old value "+oldValue+" next value "+iter.getValue());
				} while (equalValues(oldValue, iter.getValue()));

			}
			else foo.setValue(newKF.getValue());
		} else {
			sortedInsert(newKF);			
		}
		keyFramesChanged = true;
	}

	public void printState() {
		
	}
	
	public void sortedInsert(KeyFrame<T> kf)	{
		int n = keyFrames.size();
		int i = 0;
		for (i = 0; i<n; ++i)	{
			if (keyFrames.get(i).getTime() > kf.getTime()) break;
		}
		if (i == 0) {
			keyFrames.add(0, kf);
			return;
		}
		keyFrames.add(i, kf);
	}

	protected boolean keyFramesChanged() {
		if (keyFramesChanged) return true;
		for (KeyFrame kf : keyFrames)	{
			if (kf.getTimeDescriptor().lastModified > lastModified) 
				{keyFramesChanged = true; 
				lastModified = System.currentTimeMillis();
				return true;}
		}
		return false;
	}

	protected int currentSegment;
	protected boolean valueIsSet = false;	// hacky way to handle subclass recognition that superclass has done it
	protected KeyFrame<T> previous;
	protected KeyFrame<T> next;
	/**
	 * Implementation of this method is a bit tricky, since every subclass
	 * should call it when it first, then do their class-specific work.
	 * But then the subclasses have to remember to call the method
	 * {@link KeyFrameAnimatedDelegate#propagateCurrentValue(Object)} on the
	 * <i>delegate</i> field at the end of their subclass method.
	 */
	public void setValueAtTime(double t)	{
//		System.err.println(getName()+"1 setting value at time "+t);
//		System.err.println(getName()+"  Using interpolation type "+interp);
		if (keyFrames.size() == 0) {valueIsSet = true; return;}
		currentSegment = keyFrames.getSegmentAtTime(t);
		valueIsSet = false;
		// TODO these calculations should take into account the value of boundaryMode
		if (currentSegment == -1) {
			currentValue =  copyFromTo(keyFrames.getFirst().getValue(), getNewInstance());
			valueIsSet = true;
		}
		else if (currentSegment == keyFrames.size()-1) {
			currentValue =  copyFromTo(keyFrames.getLast().getValue(), getNewInstance());
			valueIsSet = true;
		} else {
			previous = keyFrames.get(currentSegment);
			next = keyFrames.get(currentSegment+1);
		}
		if (valueIsSet) return;
		if (timeMapper != null) remappedTime = timeMapper.remapTime(t, 
				previous.getTime(), 
				next.getTime());
		else remappedTime = t;
		return;
	}
	
	public List<KeyFrame<T>> getKeyFrames() {
		return keyFrames;
	}

	public void setInterpolationType(InterpolationTypes t)	{
		interp = t;
	}
	
	public InterpolationTypes getInterpolationType()	{
		return interp;
	}

	public BoundaryModes getBoundaryMode() {
		return boundaryMode;
	}

	public void setBoundaryMode(BoundaryModes modes) {
		this.boundaryMode = modes;
	}
	
	public String getName() {
		return name;
	}

	public void setName(String name) {
		this.name = name;
	}

	public boolean isWritable() {
		return writable;
	}

	public void setWritable(boolean b) {
		writable = b;
	}

	public void setFromArchive(KeyFrameAnimated<T> archived) {
		if ( !(archived instanceof SimpleKeyFrameAnimated))
			throw new IllegalArgumentException("Wrong class");
		SimpleKeyFrameAnimated<T> arch = (SimpleKeyFrameAnimated<T>) archived;
		keyFrames.clear();
		System.err.println("SKFA: Setting from archive "+archived.getName());
		List<KeyFrame<T>> archivedKF = arch.getKeyFrames();
		for (KeyFrame<T> kf : archivedKF) {
			addKeyFrame(kf);
		}
		setBoundaryMode(arch.getBoundaryMode());
		setInterpolationType(arch.getInterpolationType());
	}

	public TimeMapper getTimeMapper() {
		return timeMapper;
	}

	public void setTimeMapper(TimeMapper timeMapper) {
		this.timeMapper = timeMapper;
	}

	public void endAnimation() {
	}

	public void startAnimation() {
	}

	@Override
	public boolean isGivesWay() {
		return givesWay;
	}

	@Override
	public void setGivesWay(boolean b) {
		givesWay = b;
	}



}
