/*
 * Created on Aug 25, 2004
 *
 */
package charlesgunn.anim.gui;

import java.awt.BorderLayout;
import java.awt.Component;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.File;
import java.io.IOException;
import java.util.List;
import java.util.Vector;

import javax.swing.AbstractAction;
import javax.swing.Action;
import javax.swing.BorderFactory;
import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JCheckBox;
import javax.swing.JComboBox;
import javax.swing.JFileChooser;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JSeparator;
import javax.swing.JTabbedPane;
import javax.swing.JTextField;
import javax.swing.SwingConstants;
import javax.swing.Timer;
import javax.swing.border.TitledBorder;

import charlesgunn.anim.core.KeyFrame;
import charlesgunn.anim.core.SortedKeyFrameList;
import charlesgunn.anim.core.TimeDescriptor;
import charlesgunn.anim.io.ImportExport;
import charlesgunn.anim.tutorial.AnimatedJRViewer;
import charlesgunn.anim.util.AnimationUtility;
import charlesgunn.anim.util.TextSlider;
import charlesgunn.anim.util.AnimationUtility.PlaybackModes;
import de.jreality.util.Input;
import de.jreality.util.Secure;

/**
 * This is a GUI for managing animation tasks for classes in the package
 * <i>charlesgunn.anim</i>.  It consists of two parts, one for key frame editing
 * and another for playback and recording.  Changes to the GUI generate events of
 * type {@link AnimationPanelEvent}.  These events are propagated to listeners who
 * implement the interface {@link AnimationPanelListener}.
 * <p>
 * The key frame part of the GUI allows the user to insert new key frames with the <i>insert</i> button.
 * If the current key frame is not the last one, then new key frames are inserted halfway between
 * the current and the next one. If it is the last one, then it is inserted 1 second after the
 * current (last) one.  Once key frames
 * have been defined, the user can step through them back and forwards. He can re-define the
 * value of an existing key frame by using the <i>save</i> button, or remove it by using the 
 * <i>delete</i> button. 
 * <p>
 * The AnimationPanel keeps a list of the {@link KeyFrame} instances, without any associated values.
 * As the GUI is manipulated, these changes are propagated to the listeners, who
 * typically manage lists of key frame animated variables, which will have the same
 * {@link KeyFrame} instances as keys, associated to genuine values.
 * <p>
 * The playback/recording part of the GUI allows the user to play and pause the animation. There
 * is a slider to control the speed of the playback.  Note: this is a logarithmic slider, so its
 * range is much wider than a usual linear one. There is also a recording check-box; when this
 * is activated when the play button is pressed, then each frame will also be recorded as a
 * tiff image. 
 * <p>
 * The <i>File</i> menu allows the user to save and read back the key frame animations settings he has so 
 * laboriously (perhaps) created.  Use the "Save" and "Open..." menu items.  
 * <p>
 * The <i>Preferences</i> menu item will pop up a window where recording preferences can be edited.  These will
 * be saved and read back in with the other animation variables.
 * TODO:  	Add inspection of animated variables to allow turning them on/off selectively for different
 *  	key frames. 
 * <p>
 * See the example application
 * {@link AnimatedJRViewer} for example how to use/implement all the features described above.
 * <p>
 * @author Charles Gunn
 *
 */
public class AnimationPanel  {
	
	protected SortedKeyFrameList<Object> keyFrames = new SortedKeyFrameList<Object>();
	protected int inspectedKeyFrame = 0;
	protected KeyFrame<Object> currentKeyFrame;
	private double currentTime;
	protected int totalTicks;
	protected double totalSeconds;
	protected double playbackFactor = 1.0;
	public String stem = "spiral";
	boolean loopPlay = false, recording = false, isPaused = true;
	PlaybackModes playbackMode = PlaybackModes.NORMAL;
	protected RecordingPreferences recordPrefs;
	protected String resourceDir = Secure.getProperty("animation.readDirectory", Secure.getProperty("user.home")+"/Documents/animations/");
	public double globalPlaybackFactor = 1;
	transient protected AnimationPanelRecordListener aprl;
	protected List<AnimationPanelListener> listeners = new Vector<AnimationPanelListener>();
	
	transient JFrame prefFrame;
	transient JTabbedPane prefTabs;
	transient JPanel thePanel;
	transient JButton playButton, resetButton;
	transient PlayAction playAction;
	transient ResetAction resetAction;
	transient protected Timer playbackTimer;
	transient final JTextField keyNo, 
		timeValue, 
		durationValue; //playtime
	transient public File saveAnimationFile, readAnimationFile;
	transient protected double dcurrentTick;

	transient static Object dummyObject = new Object();
	transient private TextSlider pbspeedSlider, timeSlider;
	transient int frameCount = 0;
	transient double reversed = 1;
	transient Box timeSliderBox = Box.createHorizontalBox();
	transient private JScrollPane recordPrefsSP;
	transient int fps = 30;
	transient long systemTime;
	public AnimationPanel()	{
		
		prefFrame = new JFrame("Preferences");
		prefTabs = new JTabbedPane();
		prefFrame.getContentPane().add(prefTabs);
		recordPrefs = new RecordingPreferences();
		setRecordPrefs(recordPrefs);
		prefFrame.validate();
		prefFrame.pack();
		
		playbackTimer = new Timer(1000/recordPrefs.getFps(), new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				tick();
				// TODO do this via a listener
				if (fps != recordPrefs.getFps()) {
					fps = recordPrefs.getFps();
					int timerInterval = 1000/fps;
					playbackTimer.setDelay(timerInterval);
				}
			}
		});
		
		thePanel = new JPanel();
		thePanel.setName("Animation Panel");
		Box col = Box.createVerticalBox();

		JMenuBar mb = new JMenuBar();
		JMenu testM = new JMenu("File");
		JMenuItem jcb = new JMenuItem("Open...");
		testM.add(jcb);
		jcb.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				selectOpenFile();
				read();
			}
		});
		jcb = new JMenuItem("Save");
		testM.add(jcb);
		jcb.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				export();
			}
		});
		jcb = new JMenuItem("Save As...");
		testM.add(jcb);
		jcb.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				selectSaveFile();
				export();
			}
		});
		jcb = new JMenuItem("Clear");
		testM.add(jcb);
		jcb.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				clearKeyFrames();
			}
		});
		mb.add(testM);
		mb.add(new JSeparator());
		jcb = new JMenuItem("Preferences...");
		testM.add(jcb);
		jcb.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				prefFrame.setVisible(true);
			}
		});
		
		Box hbox = Box.createHorizontalBox();
		hbox.add(mb);
		hbox.add(Box.createHorizontalGlue());
		col.add(hbox);
		
		Box panel = Box.createHorizontalBox();
		TitledBorder title = BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), "Key frames");
		panel.setBorder(title);
		
		int strutwidth = 10;
		JLabel label = new JLabel(" # = ", SwingConstants.RIGHT);
		panel.add(label);
		panel.add(Box.createHorizontalStrut(5));
		keyNo = new JTextField(4);
		keyNo.setText("0");
		keyNo.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				int value = Integer.parseInt(keyNo.getText());
				if (value < 0 || value >= getKeyFrames().size()) {
					keyNo.setText(Integer.toString(inspectedKeyFrame)); 
					return; 
				}
				inspectedKeyFrame = value;
				inspectKeyFrame();
			}
		});
		panel.add(keyNo);
		panel.add(Box.createHorizontalStrut(strutwidth));

		panel.add(new JButton(new RetreatAction()));
		panel.add(Box.createHorizontalStrut(strutwidth));
		panel.add(new JButton(new AdvanceAction()));
		panel.add(Box.createHorizontalStrut(strutwidth));
		
		label = new JLabel(" t = ", SwingConstants.RIGHT);
		panel.add(label);
		
		timeValue = new JTextField( 8);
		timeValue.setText("0");
		timeValue.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				double value = Double.parseDouble(timeValue.getText());
				int previous = inspectedKeyFrame - 1;
				double thisTime = getKeyFrames().get(inspectedKeyFrame).getTime();
				double prevTime = 0.0;
				if (previous >= 0)
					prevTime = getKeyFrames().get(previous).getTime();
				int next = inspectedKeyFrame + 1;
				boolean atEnd = false;
				double nextTime = 0.0;
				if (next >= getKeyFrames().size()) atEnd = true;
				else nextTime = getKeyFrames().get(next).getTime();
				if ((value < prevTime) || (!atEnd && value > nextTime))	{
//					LoggingSystem.getLogger(AnimationPanel.this).warning("Invalid time");
					timeValue.setText(String.format("%6.4g", thisTime));
					return;
				}
				getKeyFrames().get(inspectedKeyFrame).setTime(value);
				setCurrentTime(getKeyFrames().get(inspectedKeyFrame).getTime());
//				System.err.println("Setting kf "+inspectedKeyFrame+" time to "+value);
			}
		});
		panel.add(timeValue);
		panel.add(Box.createHorizontalStrut(strutwidth));
		
		label = new JLabel(" dt = ", SwingConstants.RIGHT);
		panel.add(label);
		
		durationValue = new JTextField( 8);
		durationValue.setText("0");
		durationValue.addActionListener( new ActionListener() {
			public void actionPerformed(ActionEvent e)	{
				double value = Double.parseDouble(durationValue.getText());
				int next = inspectedKeyFrame + 1;
				if (next >= getKeyFrames().size()) return;
				int n = getKeyFrames().size();
				double thisTime = getKeyFrames().get(inspectedKeyFrame).getTime();
				double nextTime = getKeyFrames().get(next).getTime();
				double dt = value - (nextTime - thisTime);
				for (int i = next; i< n; ++i)	{
					double foo = keyFrames.get(i).getTime();
					keyFrames.get(i).setTime(foo + dt);
				}
//				for (AnimationPanelListener apl : listeners)
//					apl.printState();
				updateTimeSlider();
//				theCurve.setOutOfDate(true);
			}
		});
		panel.add(durationValue);
		panel.add(Box.createHorizontalStrut(strutwidth));
		
		panel.add(new JButton(new SaveKeyAction()));
		panel.add(Box.createHorizontalStrut(strutwidth));
		panel.add(new JButton(new InsertKeyAction()));
		panel.add(Box.createHorizontalStrut(strutwidth));
		panel.add(new JButton(new DeleteKeyAction()));
		
		//getContentPane().add(panel, BorderLayout.NORTH);
		col.add(Box.createVerticalStrut(10));
		col.add(panel);		
		
		Box holder = Box.createVerticalBox();
		col.add(holder);
		title = BorderFactory.createTitledBorder(BorderFactory.createEtchedBorder(), 
				"Playback");
		holder.setBorder(title);

		panel = Box.createHorizontalBox();
		holder.add(panel);

		final JCheckBox record = new JCheckBox("Recording");
		panel.add(record);
		panel.add(Box.createHorizontalStrut(strutwidth));
		record.addActionListener(new ActionListener() {

			public void actionPerformed(ActionEvent e) {
				boolean foo = record.isSelected();
				setRecording(foo);
			}
			
		});

		playAction = new PlayAction();
		playButton = new JButton( playAction);
		panel.add(playButton);
		panel.add(Box.createHorizontalStrut(strutwidth));
		
		resetAction = new ResetAction();
		resetButton = new JButton( resetAction);
		panel.add(resetButton);
		panel.add(Box.createHorizontalStrut(strutwidth));
		
//		label = new JLabel(" time ", SwingConstants.RIGHT);
//		panel.add(label);
//
//		playtime = new JTextField( 8);
//		playtime.setText("0");
//		playtime.setEnabled(false);
//		panel.add(playtime);
		
		final JLabel methodLabel = new JLabel("Playback mode:");
		String[] methods = new String[3];
		AnimationUtility.playbackModeNames.keySet().toArray(methods);
		final JComboBox method = new JComboBox(methods);
		method.addActionListener(new ActionListener() {
			public void actionPerformed(ActionEvent e) {
				setPlaybackMode(AnimationUtility.playbackModeNames.get(method.getSelectedItem()));
			}
		});
		for (int i =0; i<3; ++i)
			if (methods[i].equals("normal")) method.setSelectedIndex(i);
		
//		method.setPreferredSize(new Dimension(75, 25));
//		method.setMaximumSize(method.getPreferredSize());
		method.setAlignmentX(Component.LEFT_ALIGNMENT);
		panel.add(Box.createHorizontalStrut(strutwidth));
		panel.add(methodLabel);
		panel.add(Box.createHorizontalStrut(strutwidth));
		panel.add(method);
		panel.add(Box.createHorizontalGlue());
		
		holder.add(Box.createVerticalStrut(strutwidth));
		timeSliderBox = Box.createHorizontalBox();
		updateTimeSlider();
		holder.add(timeSliderBox);
		
		holder.add(Box.createVerticalStrut(strutwidth));
		panel = Box.createHorizontalBox();
		holder.add(panel);
		pbspeedSlider = new TextSlider.DoubleLog("playback speed",SwingConstants.HORIZONTAL,0.01,5.0,globalPlaybackFactor);
		pbspeedSlider.addActionListener(new ActionListener()	{
			public void actionPerformed(ActionEvent e)	{
				setPlaybackFactor(pbspeedSlider.getValue().doubleValue());
			}
		});
		setPlaybackFactor(globalPlaybackFactor);
		panel.add(pbspeedSlider);
		
		col.add(Box.createVerticalStrut(strutwidth));

		thePanel.add(col, BorderLayout.CENTER);

	}
	
	protected void updateTimeSlider()	{
		timeSliderBox.removeAll();
		double min = getKeyFrames().getTmin();
		double max = getKeyFrames().getTmax();
		timeSlider = new TextSlider.Double("time ",
				SwingConstants.HORIZONTAL,min, max, getCurrentTime());
		timeSlider.addActionListener(new ActionListener()	{
			public void actionPerformed(ActionEvent e)	{
				if (e.getID() != 0) return;
				if (!isPaused) return;
				setCurrentTime(timeSlider.getValue().doubleValue());
				updateTime();
			}
		});
		timeSliderBox.add(timeSlider);
//		timeSliderBox.validate();
		thePanel.validate();
	}
	protected void clearKeyFrames() {
		int n = keyFrames.size();
		for (int i =n-1; i>0; --i)	keyFrames.remove(i);
	}

	protected void selectOpenFile() {
		JFileChooser fc = new JFileChooser(resourceDir);
		//JOGLConfiguration.theLog.log(Level.FINE,"FCI resource dir is: "+resourceDir);
		int result = fc.showOpenDialog(null);
		readAnimationFile = null;
		if (result == JFileChooser.APPROVE_OPTION)	{
			readAnimationFile = fc.getSelectedFile();
		} else {
//			JOGLConfiguration.theLog.log(Level.FINE,"Unable to open file");
		}
	}
	
	protected  void read()	{
		if (readAnimationFile == null) selectOpenFile();
		// if that didn't work ...
		if (readAnimationFile == null)  {
			System.err.println("no read file specified");
			return;
		}
		ImportExport.readInto(this, readAnimationFile);
		saveAnimationFile = readAnimationFile;
	}

	public void read(String URL)	{
		Input input = null;
		try {
			input = Input.getInput(URL);
		} catch (IOException e) {
			e.printStackTrace();
		}
		read(input);
	}
	public void read(Input input)	{
		ImportExport.readInto(this, input.getInputStream());
	}

	protected void selectSaveFile() {
		JFileChooser fc = new JFileChooser(resourceDir);
		//JOGLConfiguration.theLog.log(Level.FINE,"FCI resource dir is: "+resourceDir);
		int result = fc.showSaveDialog(null);
//		saveAnimationFile = null;
		if (result == JFileChooser.APPROVE_OPTION)	{
			saveAnimationFile = fc.getSelectedFile();
		} else {
//			JOGLConfiguration.theLog.log(Level.FINE,"Unable to open file");
		}
	}
	
	protected  void export()	{
		if (saveAnimationFile == null) selectSaveFile();
		// if that didn't work ...
		if (saveAnimationFile == null)
			throw new NullPointerException("no save file specified");
		ImportExport io = new ImportExport();
		io.write(this, saveAnimationFile);
	}

	public JPanel getPanel()	{
		return thePanel;
	}
	
	public void addAnimationPanelListener(AnimationPanelListener apl)	{
		getListeners().add(apl);
	}
	
	public void removeAnimationPanelListener(AnimationPanelListener apl)	{
		getListeners().remove(apl);
	}
	
	private void fireEvent(AnimationPanelEvent ape) {
		for (AnimationPanelListener apl : getListeners())	{
			apl.actionPerformed(ape);
		}
	}
	protected void setFrameRate(int f) {		
	}

	private void inspectKeyFrame() {
		if (getKeyFrames().size() == 0) return;
		if (inspectedKeyFrame >= keyFrames.size()) inspectedKeyFrame = keyFrames.size()-1;
		currentKeyFrame = keyFrames.get(inspectedKeyFrame);
		setCurrentTime(currentKeyFrame.getTime());
		timeSlider.setValue(currentTime);
		dcurrentTick = tickFromTime(currentTime);
		updateKeyFrame();
	}

	private void updateKeyFrame() {
		if (getKeyFrames().size() == 0) return;
		if (inspectedKeyFrame<0) 
			inspectedKeyFrame = 0;
		if (inspectedKeyFrame>getKeyFrames().size()-1) 
			inspectedKeyFrame = getKeyFrames().size()-1;
		keyNo.setText(Integer.toString(inspectedKeyFrame));
		timeValue.setText(String.format("%6.4g", getKeyFrames().get(inspectedKeyFrame).getTime()));
		int next = inspectedKeyFrame+1;
		if (next >= getKeyFrames().size()) {
			durationValue.setEnabled(false);
		} else {
			durationValue.setEnabled(true);
			double t0 = getKeyFrames().get(inspectedKeyFrame).getTime();
			double t1 = getKeyFrames().get(next).getTime();
			durationValue.setText(String.format("%6.4g", t1 - t0));
		}
		currentKeyFrame = keyFrames.get(inspectedKeyFrame);
		if (timeSlider == null || getKeyFrames().getTmax() != timeSlider.getMax().doubleValue()) 
			updateTimeSlider();
		AnimationPanelEvent ape = new AnimationPanelEvent(
				AnimationPanelEvent.EventType.SET_VALUE_AT_TIME,
				AnimationPanel.this,
				 new TimeDescriptor(getCurrentTime()),
				currentKeyFrame
				);
//		System.err.println("inspected key frame = "+inspectedKeyFrame);
		fireEvent(ape);
	}

	class SaveKeyAction extends AbstractAction		{
		
		SaveKeyAction()	{
			super();
			putValue(Action.NAME, "Save");
		}
		
		public void actionPerformed(ActionEvent e) {
			if (getKeyFrames().size() == 0 || currentKeyFrame == null) {
				insertKeyFrame();
				return;
			}
			currentKeyFrame.setTime(getCurrentTime());
			AnimationPanelEvent ape = new AnimationPanelEvent(
					AnimationPanelEvent.EventType.KEY_FRAME_CHANGED,
					AnimationPanel.this,
					currentKeyFrame.getTimeDescriptor(),
					currentKeyFrame);
			fireEvent(ape);
			inspectKeyFrame();
		}


	}

	class DeleteKeyAction extends AbstractAction		{
		
		DeleteKeyAction()	{
			super();
			putValue(Action.NAME, "Delete");
		}
		
		public void actionPerformed(ActionEvent e) {
			AnimationPanelEvent ape = new AnimationPanelEvent(
					AnimationPanelEvent.EventType.KEY_FRAME_DELETED,
					AnimationPanel.this,
					currentKeyFrame.getTimeDescriptor(),
					currentKeyFrame);
			fireEvent(ape);
			getKeyFrames().remove(inspectedKeyFrame);
			inspectKeyFrame();
		}
		
	}

	class InsertKeyAction extends AbstractAction		{
		
		InsertKeyAction()	{
			super();
			putValue(Action.NAME, "Insert");
		}
		
		public void actionPerformed(ActionEvent e) {
			insertKeyFrame();
		}

		
	}

	class AdvanceAction extends AbstractAction		{
		
		AdvanceAction()	{
			super();
			putValue(Action.NAME, ">");
		}
		
		public void actionPerformed(ActionEvent e) {
			if (getKeyFrames().size() == 0) return;
			inspectedKeyFrame++;
			inspectedKeyFrame = inspectedKeyFrame % getKeyFrames().size();
			inspectKeyFrame();
		}
		
	}

	class RetreatAction extends AbstractAction		{
		
		RetreatAction()	{
			super();
			putValue(Action.NAME, "<");
		}
		
		public void actionPerformed(ActionEvent e) {
			if (getKeyFrames().size() == 0) return;
			inspectedKeyFrame--;
			if (inspectedKeyFrame < 0) inspectedKeyFrame += getKeyFrames().size();

			inspectKeyFrame();
		}
		
	}

	class SaveAllAction extends AbstractAction		{
		Component parent;
		SaveAllAction()	{
			super();
			putValue(Action.NAME, "Save As...");
		}
		
		public void actionPerformed(ActionEvent e) {
		}

	}


	class PlayAction extends AbstractAction		{
		PlayAction()	{
			super();
			putValue(Action.NAME, "Play");
		}
		
		public void actionPerformed(ActionEvent e) {
			togglePlay();
		}
	}
	
	class ResetAction extends AbstractAction		{
		ResetAction()	{
			putValue(Action.NAME, "Reset");
		}
		
		public void actionPerformed(ActionEvent e) {
			reset();
		}

	}
	
	public void reset() {
		dcurrentTick = 0.0;
		frameCount = 0;
		playbackTimer.stop();
		reversed = 1;
		setCurrentTime(0.0);
		updateTime();
	}
	private void insertKeyFrame() {
		KeyFrame<Object> kf;
		if (getKeyFrames().size() == 0)	{
			kf = new KeyFrame(new TimeDescriptor(0.0), dummyObject);
		} 
		else if (getKeyFrames().size()-1 == inspectedKeyFrame) {
			kf = new KeyFrame(new TimeDescriptor(getKeyFrames().getTmax()+1), dummyObject);
		} 
		else {
			double tm = (getKeyFrames().get(inspectedKeyFrame).getTime() + getKeyFrames().get(inspectedKeyFrame+1).getTime())*.5;
			kf = new KeyFrame(new TimeDescriptor(tm), dummyObject);
		}
		getKeyFrames().sortedInsert(kf);
		inspectedKeyFrame = getKeyFrames().indexOf(kf);
		currentKeyFrame = kf;
		setCurrentTime(kf.getTime());
		AnimationPanelEvent ape = new AnimationPanelEvent(
				AnimationPanelEvent.EventType.KEY_FRAME_ADDED,
				AnimationPanel.this,
				currentKeyFrame.getTimeDescriptor(),
				currentKeyFrame);
		fireEvent(ape);
		inspectKeyFrame();
	}
	
	public int tickFromTime(double t)	{
		return (int) (totalTicks * (t - getKeyFrames().getTmin())/(getKeyFrames().getTmax() - getKeyFrames().getTmin()));
	}
	
	public void tick()	{
		// it's important to do the last frame
//		System.err.println("in: current tick = "+dcurrentTick);
		if (dcurrentTick > totalTicks) dcurrentTick = totalTicks;
		double absoluteTime = (dcurrentTick)/totalTicks;
		double totalTime = getKeyFrames().getTmax() - getKeyFrames().getTmin();
		setCurrentTime(getKeyFrames().getTmin() + absoluteTime * totalTime);
		updateTime();
		if (dcurrentTick >= totalTicks || dcurrentTick < 0) {
			switch(playbackMode)	{
				case NORMAL:
					frameCount = (int) (dcurrentTick=0);
					if (recording) recording = false;
					isPaused = true;
					reversed = 1;
					break;
				case CYCLE:
					frameCount = (int) (dcurrentTick=0);	
					reversed = 1;
					break;
				case SHUTTLE:
					reversed = -reversed;
					dcurrentTick = (dcurrentTick >= totalTicks) ? (totalTicks-playbackFactor) : 0;
					break;
			}
			AnimationPanelEvent ape = new AnimationPanelEvent(
					AnimationPanelEvent.EventType.PLAYBACK_COMPLETED,
					AnimationPanel.this,
					currentKeyFrame.getTimeDescriptor(),
					currentKeyFrame);
			fireEvent(ape);
			updatePlayState();
		}
		// update the time
		if (recording)	{
			dcurrentTick += reversed * playbackFactor;			
		} else { 		//read the clock
			long newTime = System.currentTimeMillis();
			long dt = newTime - systemTime;
			systemTime = newTime;
			double dticks = (dt/1000.0) * fps* playbackFactor * reversed;
			dcurrentTick += dticks;
		}  // check boundary conditions
//			return;
//		System.err.println("out: current tick = "+dcurrentTick);
		frameCount++;
	}
	/**
	 * This method is called generally in playback mode
	 */
	public void updateTime()	{
		int which = getKeyFrames().getSegmentAtTime(getCurrentTime());
		if (which != inspectedKeyFrame) {
			inspectedKeyFrame = which;
			updateKeyFrame();
		}
//		System.err.println("Current time = "+currentTime+" Current kf = "+inspectedKeyFrame);
		timeSlider.setValue(currentTime);
		AnimationPanelEvent ape = new AnimationPanelEvent(
				AnimationPanelEvent.EventType.SET_VALUE_AT_TIME,
				AnimationPanel.this,
				new TimeDescriptor(currentTime),
				currentKeyFrame);
		fireEvent(ape);
//		System.err.println("time = "+currentTime);
	}
	
	public void togglePlay() {
		if (getKeyFrames().size() == 0) return;
		isPaused = !isPaused;
		updatePlayState();
	}
	
	private void updatePlayState() {
//		System.err.println("In play state paused = "+isPaused);
		if (isPaused) {
			playbackTimer.stop();
			playAction.putValue(Action.NAME, "Play");
			timeSlider.setValue(getCurrentTime());
		} else	{
			totalSeconds = getKeyFrames().getTmax() - getKeyFrames().getTmin();
			fps = recordPrefs.getFps();
			totalTicks = (int) (totalSeconds * fps);
			System.err.println("Total ticks = "+totalTicks);
			systemTime = System.currentTimeMillis();
//			dcurrentTick = tickFromTime(getCurrentTime())
			AnimationPanelEvent ape = new AnimationPanelEvent(
					AnimationPanelEvent.EventType.PLAYBACK_STARTED,
					AnimationPanel.this,
					 new TimeDescriptor(getCurrentTime()),
					currentKeyFrame
					);
//			System.err.println("inspected key frame = "+inspectedKeyFrame);
			fireEvent(ape);
			playbackTimer.start();
			playAction.putValue(Action.NAME, "Pause");
		}
	}

	public void clear() {
		inspectedKeyFrame = 0;
		inspectKeyFrame();
		keyFrames.clear();
	}
	public PlaybackModes getPlaybackMode() {
		return playbackMode;
	}
	public void setPlaybackMode(PlaybackModes mode) {
		playbackMode = mode;
	}

	public double getPlaybackFactor() {
		return playbackFactor;
	}
	
	public void setPlaybackFactor(double playbackFactor) {
		this.playbackFactor = playbackFactor;
		pbspeedSlider.setValue(playbackFactor);
	}

	public boolean isRecording() {
		return recording;
	}

	public int getCurrentFrame() {
		return frameCount;
	}

	public void setCurrentTime(double currentTime) {
		this.currentTime = currentTime;
	}

	public double getCurrentTime() {
		return currentTime;
	}

	public int getInspectedKeyFrame() {
		return inspectedKeyFrame;
	}

	public void setInspectedKeyFrame(int inspectedKeyFrame) {
		this.inspectedKeyFrame = inspectedKeyFrame;
		updateKeyFrame();
	}

	public KeyFrame<Object> getCurrentKeyFrame() {
		return currentKeyFrame;
	}

	public void setCurrentKeyFrame(KeyFrame<Object> currentKeyFrame) {
		this.currentKeyFrame = currentKeyFrame;
	}

	public int getTotalTicks() {
		return totalTicks;
	}

	public void setTotalTicks(int totalTicks) {
		this.totalTicks = totalTicks;
	}

	public double getTotalSeconds() {
		return totalSeconds;
	}

	public void setTotalSeconds(double totalSeconds) {
		this.totalSeconds = totalSeconds;
	}

	public boolean isPaused() {
		return isPaused;
	}

	public void setPaused(boolean isPaused) {
		this.isPaused = isPaused;
		updatePlayState();
	}

	public  String getResourceDir() {
		return resourceDir;
	}

	public void setResourceDir(String resourceDir) {
		this.resourceDir = resourceDir;
	}
	
	public double getGlobalPlaybackFactor() {
		return globalPlaybackFactor;
	}

	public void setGlobalPlaybackFactor(double globalPlaybackFactor) {
		this.globalPlaybackFactor = globalPlaybackFactor;
	}

	public int getFrameCount() {
		return frameCount;
	}

	public void setFrameCount(int frameCount) {
		this.frameCount = frameCount;
	}

	public JTextField getDurationValue() {
		return durationValue;
	}

	public void setRecording(boolean recording) {
		this.recording = recording;
	}

	public void setKeyFrames(SortedKeyFrameList<Object> keyFrames) {
		this.keyFrames = keyFrames;
	}

	public SortedKeyFrameList<Object> getKeyFrames() {
		return keyFrames;
	}

	public void setListeners(List<AnimationPanelListener> listeners) {
		this.listeners = listeners;
	}

	public List<AnimationPanelListener> getListeners() {
		return listeners;
	}
	
	public void setFromArchive(Object o)	{
		if ( !(o instanceof AnimationPanel)) {
			throw new IllegalArgumentException("Wrong class");
		}
		AnimationPanel archPanel = (AnimationPanel) o;
		SortedKeyFrameList<Object> kf = archPanel.getKeyFrames();
		getKeyFrames().clear();
		for (int i = 0; i<kf.size(); ++i) {
			getKeyFrames().add(kf.get(i));
		}
		if (archPanel.getListeners() != null)  {
			for (AnimationPanelListener apl : archPanel.getListeners()) {
				if (apl !=null) System.err.println("Read listener named "+apl.getName());
			}

			for (AnimationPanelListener apl : getListeners()) {
				AnimationPanelListener xxx = 
					(AnimationPanelListener) elementNamed(apl.getName(), (archPanel.getListeners()));
				if (xxx == null) {
					System.err.println("Can't find element named "+apl.getName());
					continue;
				}
				apl.setState(xxx.getState());
			}			
		}
		setInspectedKeyFrame(archPanel.getInspectedKeyFrame());
		setCurrentKeyFrame(archPanel.getCurrentKeyFrame());
		setCurrentTime(archPanel.getCurrentTime());
		setTotalTicks(archPanel.getTotalTicks());
		setTotalSeconds(archPanel.getTotalSeconds());
		setPlaybackFactor(archPanel.getPlaybackFactor());
		RecordingPreferences rp = archPanel.getRecordPrefs();
		if (rp != null)
			setRecordPrefs(archPanel.getRecordPrefs());
		updateTimeSlider();
	}
	
	public static AnimationPanelListener elementNamed(String name, List<AnimationPanelListener> list)	{
		for (AnimationPanelListener kfa :  list)	{
			if (kfa.getName().equals(name)) return kfa;
		}
		return null;
	}

	public RecordingPreferences getRecordPrefs() {
		return recordPrefs;
	}

	public void setRecordPrefs(RecordingPreferences recordPrefs) {
		this.recordPrefs = recordPrefs;
		if (recordPrefsSP != null) prefTabs.remove(recordPrefsSP);
		recordPrefsSP = new JScrollPane(recordPrefs.getComponent());
		recordPrefsSP.setName("Recording");
		prefTabs.add(recordPrefsSP);
	}

}
