package charlesgunn.anim.tutorial;

import static de.jreality.shader.CommonAttributes.DIFFUSE_COLOR;
import static de.jreality.shader.CommonAttributes.LINE_SHADER;
import static de.jreality.shader.CommonAttributes.POLYGON_SHADER;

import java.awt.Color;
import java.awt.Component;
import java.awt.Dimension;
import java.awt.GridLayout;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.io.IOException;
import java.util.List;
import java.util.Vector;

import javax.swing.Box;
import javax.swing.JButton;
import javax.swing.JColorChooser;
import javax.swing.JFrame;
import javax.swing.SwingConstants;

import charlesgunn.anim.core.Animated;
import charlesgunn.anim.core.KeyFrameAnimated;
import charlesgunn.anim.core.KeyFrameAnimatedBean;
import charlesgunn.anim.core.KeyFrameAnimatedColor;
import charlesgunn.anim.core.KeyFrameAnimatedDelegate;
import charlesgunn.anim.core.KeyFrameAnimatedDouble;
import charlesgunn.anim.core.KeyFrameAnimatedTransformation;
import charlesgunn.anim.gui.AnimationPanel;
import charlesgunn.anim.gui.AnimationPanelListenerImpl;
import charlesgunn.anim.gui.AnimationPanelRecordListener;
import charlesgunn.anim.jreality.SceneGraphAnimator;
import charlesgunn.anim.plugin.AnimationPlugin;
import charlesgunn.anim.util.AnimationUtility.InterpolationTypes;
import de.jreality.examples.CatenoidHelicoid;
import de.jreality.shader.DefaultPolygonShader;
import de.jreality.plugin.JRViewer;
import de.jreality.plugin.JRViewerUtility;
import de.jreality.plugin.JRViewer.ContentType;
import de.jreality.plugin.basic.Scene;
import de.jreality.plugin.basic.View;
import de.jreality.plugin.scene.SceneShrinkPanel;
import de.jreality.scene.Appearance;
import de.jreality.scene.Camera;
import de.jreality.scene.SceneGraphComponent;
import de.jreality.scene.SceneGraphPath;
import de.jreality.scene.SceneGraphVisitor;
import de.jreality.scene.Viewer;
import de.jreality.shader.DefaultGeometryShader;
import de.jreality.shader.ShaderUtility;
import de.jreality.shader.TwoSidePolygonShader;
import de.jreality.tools.ClickWheelCameraZoomTool;
import de.jreality.tools.DraggingTool;
import de.jreality.tools.EncompassTool;
import de.jreality.tools.FlyTool;
import de.jreality.tools.RotateTool;
import charlesgunn.anim.util.TextSlider;
import de.jreality.ui.viewerapp.ViewerApp;
import de.jreality.util.CameraUtility;
import de.jreality.util.Input;
import de.jreality.util.SceneGraphUtility;
import de.jtem.jrworkspace.plugin.PluginInfo;
import de.jtem.jrworkspace.plugin.sidecontainer.SideContainerPerspective;
import de.jtem.jrworkspace.plugin.sidecontainer.template.ShrinkPanelPlugin;

/**
 * This is a tutorial class demonstrating many of the features of the animation package 
 * applied to a simple jReality scene graph.  Invoke the main method of this class to
 * run it.
 * 
 * The application pops up a jReality {@link JRViewer} instance containing a
 * 3D scene featuring a helicoid. To display the animation panel, select the menu item
 * "Window->Top Slot".  This should display an animation panel (instance of {@link AnimationPanel}.
 * For convenience, you can drag it out of the frame and make it a separate frame.
 * <p>
 * The following features are animated:
 * <ul>
 * <li>the parameter <i>alpha</i> for controlling the homotopy between helicoid and catenoid, </li>
 * <li>the color of the tubes drawn on the edges of the faces, </li>
 * <li>the camera in the scene, and </li>
 * <li>all the transformations in the scene graph.</li>
 * </ul>
 * <p>
 * The first two animated variables (instances of {@link KeyFrameAnimatedDouble}
 * and {@link KeyFrameAnimatedColor}, resp.) can be changed by using the "Parameters" tab of the
 * navigator in the JRViewer frame (the left panel, click on "Window->Left Slot" if it's not visible).  
 * The camera can be animated for example by using
 * the scroll wheel: there is a tool hooked up to the scroll wheel which
 * controls the field of view. Other parameters can be controlled by using the controls on the
 * "Camera" menu.   
 * <p>
 * Animation of the camera illustrates use of 
 * {@link KeyFrameAnimatedBean}. The transformation animation illustrates use of
 * {@link KeyFrameAnimatedTransformation}, and hooking up all of the transformations
 * was achieved by an instance of {@link SceneGraphAnimator}, a subclass of 
 * {@link SceneGraphVisitor}.
 * <p>
 * Recommended mode of usage: begin by clicking on "Save" button in animation panel. This 
 * sets a keyframe at time 0. 
 * Change some settings until you have a reasonable picture, and click "Insert". This sets a 
 * second key frame at time 1 (second).  Change
 * some more, for example, rotate the object with the left mouse. Click insert again.  This 
 * sets a third key frame at 2 seconds.
 * <p>
 * You should now have 3 key frames. Click on play.
 * You should see an animation lasting about 3 seconds, passing through your 3 key frames.
 * Click through the key frames, make changes, and click "Save" to save the revised states (individually).
 * <p>
 * Use the slider below the play button to control speed of playback. Recording
 * is as easy as clicking the "Recording" check box before clicking "Play". Make sure
 * you've set the directory and file stem correctly before starting. Use the "File->Preferences..." panel
 * to make these changes.
 * <p>
 *
 * * @author Charles Gunn
 *
 */public class AnimatedJRViewer {

	private TextSlider alphaSlider;
	SceneGraphComponent world;
	CatenoidHelicoid cathel = new CatenoidHelicoid(40);
	Dimension dim = new Dimension(1064, 798);
	Color tubeColor = Color.white;
	private List<Animated> animated;

	public SceneGraphComponent makeWorld() {
		world = SceneGraphUtility.createFullSceneGraphComponent("world");
		world.addTool(new RotateTool() );
		world.addTool(new DraggingTool() );
		world.addTool(new EncompassTool());
		world.addTool(new ClickWheelCameraZoomTool());
		updateWorld();
		world.setGeometry(cathel);
		Appearance ap = world.getAppearance();
		DefaultGeometryShader dgs = (DefaultGeometryShader) 
			ShaderUtility.createDefaultGeometryShader(ap, true);
		TwoSidePolygonShader tsps = (TwoSidePolygonShader) dgs.createPolygonShader("twoSide");
		ap.setAttribute(POLYGON_SHADER+"name","twoSide");
		DefaultPolygonShader dps1 = (DefaultPolygonShader) tsps.createBack("default"); 
		dps1.setDiffuseColor(new Color(0,204,204));
		DefaultPolygonShader dps2 = (DefaultPolygonShader) tsps.createBack("default"); 
		dps2.setDiffuseColor( new Color(204,204,0));
		ap.setAttribute(LINE_SHADER+"."+DIFFUSE_COLOR, java.awt.Color.WHITE);

		return world;
	}
	double alpha;
	void updateWorld()	{
		cathel.setAlpha(alpha);
	}
	
	private void setupAnimation(AnimationPanel ap)	{
		animated = new Vector<Animated>();
		
		// Here's how to manually animate the scene graph, and the camera;
		// we use the AnimationPlugin() to do this for us (see doIt() below).
//		SceneGraphAnimator sga = new SceneGraphAnimator(viewer.getSceneRoot());
//		// we animate the camera by hand below to show how to use the "bean" feature"
//		sga.setAnimateCamera(false);
//		sga.setDefaultInterp(InterpolationTypes.CUBIC_HERMITE);
//		sga.setName("sga");
//		sga.init();
//		animated.add(sga);
		
//		Camera c = CameraUtility.getCamera(viewer);
//		KeyFrameAnimatedBean<Camera> ac = new KeyFrameAnimatedBean<Camera>(c);
//		ac.setName("cameraBean");
//		animated.add(ac);
		
		// animate the color of the tubes on the world
		KeyFrameAnimatedDelegate<Color> dc = new KeyFrameAnimatedDelegate<Color> () {
			Appearance ap = world.getAppearance();
			public void propagateCurrentValue(Color t) {
				ap.setAttribute("lineShader.polygonShader.diffuseColor", t);
				viewer.renderAsync();
			}

			public Color gatherCurrentValue(Color t) {
				Object foo = ap.getAttribute("lineShader.polygonShader.diffuseColor", Color.class);
				if (foo instanceof Color)	{
					System.err.println("Gathering color "+((Color)foo));

					return ((Color) foo);
				}
				return Color.white;  // signal error!
			}
			
		};
		KeyFrameAnimatedColor lineCA = new KeyFrameAnimatedColor(dc);
		lineCA.setName("lineCA");
		animated.add(lineCA);
		
		// animate geometric deformation
		KeyFrameAnimatedDelegate<Double> dd = new KeyFrameAnimatedDelegate<Double> () {

			public void propagateCurrentValue(Double t) {
				alpha = t;
				alphaSlider.setValue(t);
				updateWorld();
				viewer.renderAsync();
			}

			public Double gatherCurrentValue(Double t) {
				return alpha;
			}
			
		};
		KeyFrameAnimatedDouble animAlpha = new KeyFrameAnimatedDouble(dd );
		animAlpha.setName("animAlpha");
		animated.add(animAlpha);
		
		AnimationPanelListenerImpl apl = new AnimationPanelListenerImpl(viewer, "tutorial demo");
		apl.setAnimated(animated);
		ap.addAnimationPanelListener(apl);
		AnimationPanelRecordListener aprl = new AnimationPanelRecordListener(viewer, "record");
		ap.addAnimationPanelListener(aprl);
	}
	
	public Component getInspector() {	
		Box inspectionPanel =  Box.createVerticalBox();
		alphaSlider = new TextSlider.Double("alpha",SwingConstants.HORIZONTAL,0.0,Math.PI*2,alpha);
		alphaSlider.addActionListener(new ActionListener()	{
			public void actionPerformed(ActionEvent e)	{
				alpha = alphaSlider.getValue().doubleValue();
				updateWorld();
				viewer.renderAsync();
			}
		});
		inspectionPanel.add(alphaSlider);
		JButton tubeColorB = new JButton("tubeColor ");
		tubeColorB.setBackground(tubeColor);
		tubeColorB.addActionListener(new ActionListener()	{
			public void actionPerformed(ActionEvent e)	{
				Color color = JColorChooser.showDialog(
						(Component) viewer.getViewingComponent(), "Select color ",  null);
				if (color != null) {
					world.getAppearance().setAttribute("lineShader.polygonShader.diffuseColor", color);
				}
			}
		});
		inspectionPanel.add(tubeColorB);
		inspectionPanel.setName("Parameters");
		return inspectionPanel;
	}


	AnimationPanel ap;
	Viewer viewer;
	SceneGraphPath avatarPath;
	void doIt()	{
	    JRViewer jrv = new JRViewer();
	    jrv.addBasicUI();
	    jrv.addContentSupport(ContentType.Raw);
	    
	    ShrinkPanelPlugin spp = createShrinkPanel(getInspector(), "Animation tutorial");
	    jrv.registerPlugin(spp);
	    AnimationPlugin plugin = new AnimationPlugin();
	    jrv.registerPlugin(plugin);
	    jrv.setContent(makeWorld());
	    jrv.startup();
	   
	    plugin.setAnimateCamera(true);
	    plugin.setAnimateSceneGraph(true);
	    
	    viewer = jrv.getViewer();
	    viewer.getSceneRoot().addTool(new EncompassTool());
	    setupAnimation(plugin.getAnimationPanel());
	    Scene scene = jrv.getPlugin(Scene.class); 
	    avatarPath = scene.getAvatarPath();
	    avatarPath.getLastComponent().addTool(new FlyTool());
		try {
			plugin.getAnimationPanel().read(
					new Input(this.getClass().getResource("animationTutorial-01.xml")));
		} catch (IOException e) {
			// TODO Auto-generated catch block
			e.printStackTrace();
		}

	}
	
	// wrap an instance of Component as a plugin
	public static SceneShrinkPanel createShrinkPanel( final Component c, final String title) {
		SceneShrinkPanel p = new SceneShrinkPanel() {
			
			{
				GridLayout gl = new GridLayout();
				gl.setRows(1);
				setInitialPosition(SHRINKER_LEFT);
				shrinkPanel.setName(title);
				shrinkPanel.setLayout(gl);
				shrinkPanel.add(c);
			}
			
			@Override
			public Class<? extends SideContainerPerspective> getPerspectivePluginClass() {
				return View.class;
			}

			@Override
			public PluginInfo getPluginInfo() {
				return new PluginInfo(title);
			}
		};
		return p;

	}


	public static void main(String[] args) {
		new AnimatedJRViewer().doIt();
	}


}
