package charlesgunn.anim.util;

import java.awt.Color;
import java.awt.geom.Rectangle2D;
import java.util.HashMap;
import java.util.List;
import java.util.Vector;

import charlesgunn.anim.core.Animated;
import charlesgunn.anim.core.KeyFrameAnimated;
import charlesgunn.anim.core.KeyFrameAnimatedBean;
import charlesgunn.anim.gui.AnimationPanel;
import charlesgunn.anim.gui.AnimationPanelListenerImpl;
import charlesgunn.anim.jreality.SceneGraphAnimator;
import de.jreality.math.FactoredMatrix;
import de.jreality.math.Pn;
import de.jreality.math.Quaternion;
import de.jreality.math.Rn;
import de.jreality.scene.Camera;
import de.jreality.scene.Viewer;
import de.jreality.util.CameraUtility;

/**
 * Static methods for implementing various interpolations on various primitive types of data.
 * @author Charles Gunn
 *
 */
public class AnimationUtility {

	public enum InterpolationTypes  {CONSTANT, LINEAR, CUBIC_HERMITE, CUBIC_BSPLINE};

	public enum BoundaryModes {CLAMP, REPEAT, EXTRAPOLATE};
	
	public enum PlaybackModes {NORMAL, CYCLE, SHUTTLE};
	public static HashMap<String, PlaybackModes> playbackModeNames = new HashMap<String, PlaybackModes>();
	static {
		playbackModeNames.put("normal", PlaybackModes.NORMAL);
		playbackModeNames.put("cycle", PlaybackModes.CYCLE );
		playbackModeNames.put("shuttle",PlaybackModes.SHUTTLE);
	}
	
	private AnimationUtility()	{
		super();
	}
	
	public static double linearInterpolation(double t, double minin, double maxin, double minout, double maxout)	{
		if (t <= minin) return minout;
		if (t >= maxin) return maxout;
		double in = (t-minin)/(maxin-minin);
		return minout + in*(maxout-minout);
	}
	
	public static double hermiteInterpolation(double t, double minin, double maxin, double minout, double maxout)	{
		if (t <= minin) return minout;
		if (t >= maxin) return maxout;
		double in = (t-minin)/(maxin-minin);
		double out = -2*in*in*in + 3*in*in;
		return minout + out*(maxout-minout);
	}
	
	public static Color linearInterpolation(double t, double minin, double maxin, Color minout, Color maxout)	{
		if (t <= minin) return minout;
		if (t >= maxin) return maxout;
		double in = (t-minin)/(maxin-minin);
		return linearInterpolation(minout, maxout, in);
	}
	
	public static Color hermiteInterpolation(double t, double minin, double maxin, Color minout, Color maxout)	{
		if (t <= minin) return minout;
		if (t >= maxin) return maxout;
		double in = (t-minin)/(maxin-minin);
		double out = -2*in*in*in + 3*in*in;
		return linearInterpolation(minout, maxout, out);
	}

	public static double[] linearInterpolation(double t, double minin, double maxin, double[] minout, double[] maxout)	{
		return linearInterpolation(null, t, minin, maxin, minout, maxout);
	}
	public static double[] linearInterpolation(double[] dst, double t, double minin, double maxin, double[] minout, double[] maxout)	{
		if (dst == null) dst = new double[minout.length];
		double in = getStandardTime(t, minin, maxin);
//		double[] result = new double[minout.length];
		for (int i = 0; i<minout.length; ++i)	{
			dst[i] = minout[i] + in*(maxout[i]-minout[i]);
		}
		return dst;
	}
	
	public static double[][] linearInterpolation(double[][] dst, double t, double minin, double maxin, double[][] minout, double[][] maxout)	{
		if (dst == null) dst = new double[minout.length][minout[0].length];
		double in = getStandardTime(t, minin, maxin);
//		double[] result = new double[minout.length];
		for (int i = 0; i<minout.length; ++i)	{
			dst[i] = linearInterpolation(dst[i],t, minin, maxin, minout[i], maxout[i]);
		}
		return dst;
	}
	
	
	public static double[] hermiteInterpolation(double t, double minin, double maxin, double[] minout, double[] maxout)	{
		return hermiteInterpolation(null, t, minin, maxin, minout, maxout);
	}
	
	public static double[] hermiteInterpolation(double[] result, double t, double minin, double maxin, double[] minout, double[] maxout)	{
		if (t <= minin) return minout;
		if (t >= maxin) return maxout;
		if (result == null) result = new double[minout.length];
		for (int i = 0; i<minout.length; ++i)	{
			double in = (t-minin)/(maxin-minin);
			double out = -2*in*in*in + 3*in*in;
			result[i] = minout[i] + out*(maxout[i]-minout[i]);
		}
		return result;
	}

	public static double[][] hermiteInterpolation(double t, double minin, double maxin, double[][] minout, double[][] maxout)	{
		return hermiteInterpolation(null, t, minin, maxin, minout, maxout);
	}
	public static double[][] hermiteInterpolation(double[][] result, double t, double minin, double maxin, double[][] minout, double[][] maxout)	{
		if (t <= minin) return minout;
		if (t >= maxin) return maxout;
		if (result == null || result.length != minout.length || result[0].length != minout[0].length)
			result = new double[minout.length][minout[0].length];
		for (int i = 0; i<minout.length; ++i)	{
			result[i] = hermiteInterpolation(t, minin, maxin, minout[i], maxout[i]);
		}
		return result;
	}
	

	public static double linearInterpolation(double d1, double d2, double s)	{
		if (s <= 0) return d1;
		if (s >= 1) return d2;
		return (1-s)*d1 + s * d2;
	}
	
	public static double hermiteInterpolation(double d1, double d2, double s)	{
		if (s <= 0) return d1;
		if (s >= 1) return d2;
		double out = -2*s*s*s + 3*s*s;
		return (1-out)*d1 + out * d2;
	}
	
	public static Rectangle2D linearInterpolation(double t, double t0, double t1, Rectangle2D r1, Rectangle2D r2)	{
		Rectangle2D dst = new Rectangle2D.Double();
		if (t <= t0) return r1;
		if (t >= t1) return r2;
		double in = (t-t0)/(t1-t0);
		double x = linearInterpolation(r1.getX(), r2.getX(), in);
		double y = linearInterpolation(r1.getY(), r2.getY(), in);
		double w = linearInterpolation(r1.getWidth(), r2.getWidth(), in);
		double h = linearInterpolation(r1.getHeight(), r2.getHeight(), in);
		dst.setRect(x,y,w,h);
		return dst;
	}

	public static Rectangle2D hermiteInterpolation(double t, double t0, double t1, Rectangle2D r1, Rectangle2D r2)	{
		Rectangle2D dst = new Rectangle2D.Double();
		if (t <= t0) return r1;
		if (t >= t1) return r2;
		double in = (t-t0)/(t1-t0);
		double x = hermiteInterpolation(r1.getX(), r2.getX(), in);
		double y = hermiteInterpolation(r1.getY(), r2.getY(), in);
		double w = hermiteInterpolation(r1.getWidth(), r2.getWidth(), in);
		double h = hermiteInterpolation(r1.getHeight(), r2.getHeight(), in);
		dst.setRect(x,y,w,h);
		return dst;
	}
	

	public static FactoredMatrix linearInterpolation(FactoredMatrix dst, double t, double tmin, double tmax, FactoredMatrix t1, FactoredMatrix t2) {
		return linearInterpolation(dst, t1, t2, (t-tmin)/(tmax-tmin));
	}
	
	public static FactoredMatrix linearInterpolation(FactoredMatrix dst, FactoredMatrix t1, FactoredMatrix t2, double s ) {
		if (dst == null) dst = new FactoredMatrix(t1.getMetric());
		int sig = t1.getMetric();
		if (sig != t2.getMetric() || sig != dst.getMetric())	{
			throw new IllegalArgumentException("Differing metrics ");
		}
		
		double[] trans1 = t1.getTranslation();
		double[] trans2 = t2.getTranslation();
		double[] trans = Pn.linearInterpolation(null,trans1, trans2, s, sig);
		
//		double[] center1 = t1.getCenter();
//		double[] center2 = t2.getCenter();
//		double[] center = null;
//		if (center1 != null && center2 != null)
//			center = Pn.linearInterpolation(null,center1, center2, s, sig);
//		
		double[] stretch1 = t1.getStretch();
		double[] stretch2 = t2.getStretch();
		double[] stretch = Pn.linearInterpolation(null,stretch1, stretch2, s, Pn.EUCLIDEAN);
		
		Quaternion rot1 = t1.getRotationQuaternion();
		Quaternion rot2 = t2.getRotationQuaternion();
		Quaternion rot = Quaternion.linearInterpolation(null, rot1, rot2, s);

		dst.setTranslation(trans);
//		if (center != null) dst.setCenter(center);
		dst.setStretch(stretch);
		dst.setRotation(rot);
		return dst;
	}
	
	/**
	 * * @param c1
	 * @param c2
	 * @param d
	 * @return (1-d) c1 + d * c2
	 *
	 */
	static float[] fc1 = new float[4];
	static float[] fc2 = new float[4];
	public static Color linearInterpolation(Color c1, Color c2, double d) {
		c1.getRGBComponents(fc1);
		c2.getRGBComponents(fc2);
		float fd = (float) d;
		return new Color((1-fd)*fc1[0]+fd*fc2[0],(1-fd)*fc1[1]+fd*fc2[1],(1-fd)*fc1[2]+fd*fc2[2], (1-fd)*fc1[3]+fd*fc2[3]);
	}

//	public static de.jreality.util.Color linearInterpolation(de.jreality.util.Color c1, de.jreality.util.Color c2, double d) {
//		c1.getRGBComponents(fc1);
//		c2.getRGBComponents(fc2);
//		float fd = (float) d;
//		return new de.jreality.util.Color((1-fd)*fc1[0]+fd*fc2[0],(1-fd)*fc1[1]+fd*fc2[1],(1-fd)*fc1[2]+fd*fc2[2], (1-fd)*fc1[3]+fd*fc2[3]);
//	}

	public static double[] matrixFromRectangle(double[] dst, Rectangle2D r)	{
		if (dst == null) dst = Rn.identityMatrix(4);
		else Rn.setIdentityMatrix(dst);
		dst[0] = r.getWidth();
		dst[5] = r.getHeight();
		dst[3] = r.getX();
		dst[7] = r.getY();
//		Rn.conjugateByMatrix(dst, dst, specialTranslate);
		return dst;
	}

	public static AnimationPanel animateViewer( Viewer viewer)	{
		return animateViewer(null, viewer);
	}

	public static AnimationPanel animateViewer(AnimationPanel ap, Viewer viewer)	{
		List<Animated> animated = new Vector<Animated>();
		
		SceneGraphAnimator sga = new SceneGraphAnimator(viewer.getSceneRoot());
		sga.setAnimateCamera(false);
		sga.setName("sga");
		sga.init();
		animated.add(sga);
		
		Camera c = CameraUtility.getCamera(viewer);
		KeyFrameAnimatedBean<Camera> ac = new KeyFrameAnimatedBean<Camera>(c);
		ac.setName("cameraBean");
		animated.add(ac);
		
		AnimationPanelListenerImpl apl = new AnimationPanelListenerImpl(viewer, "application");
		apl.setAnimated(animated);
		if (ap == null) ap = new AnimationPanel();
		ap.addAnimationPanelListener(apl);

		return ap;

	}
	
	public static void resetSceneGraph(AnimationPanel ap, Viewer viewer)	{
		
	}

	public static void getStandardInfo(double[] st, int[] key, double t, double[] keySet) {
		int i = 0;
		int n = keySet.length;
		// find the pair of keys (key, key+1) such that t is contained 
		// in the half-open interval [key, key+1)
		// first find the first key frame bigger than or equal to t: it's the right hand key frame
		while ( i < n && t >= keySet[i] ) i++;
		i--;
		if (i < 0) {st[0] = 0.0; key[0] = key[1] = 0; }
		else if (i>=(n-1)) {st[0] = 1.0; key[0] = key[1] = n-1;}
		else {
			st[0] = getStandardTime(t, keySet[i], keySet[i+1]);
			key[0] = i; key[1] = i+1;
		}
//		System.err.println("standard info: t="+st[0]+"key="+key[0]+"key1="+key[1]+"\tdt="+(keySet[i+1]-keySet[i]));
	}

	public static double getStandardTime(double t, double minin, double maxin) {
		if (minin == maxin) return 0.0;
		if (t <= minin) t = minin;
		if (t >= maxin) t = maxin;
		double in = (t-minin)/(maxin-minin);
		return in;
	}

}
