/**
 *
 * This package is open source software, made available under a BSD license:
 *
 * Copyright (c) 2009, Charles Gunn
 *
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * - Redistributions of source code must retain the above copyright notice, this
 *   list of conditions and the following disclaimer.
 *
 * - Redistributions in binary form must reproduce the above copyright notice,
 *   this list of conditions and the following disclaimer in the documentation
 *   and/or other materials provided with the distribution.
 *
 * - Neither the name of jReality nor the names of its contributors nor the
 *   names of their associated organizations may be used to endorse or promote
 *   products derived from this software without specific prior written
 *   permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */


package de.jtem.discretegroup.core;

import java.util.HashMap;
import java.util.WeakHashMap;
import java.util.logging.Level;

import de.jreality.math.Matrix;
import de.jreality.math.P3;
import de.jreality.math.Pn;
import de.jreality.math.Rn;
import de.jreality.scene.Geometry;
import de.jreality.scene.SceneGraphComponent;

/**
 * This is the base class of the <code>de.jtem.discretegroup</code> package.
 * It is a class which serves primarily to represent discrete groups which
 * are:
 * <ul>
 * <li>represented by elements of PGL(4), 4x4 real matrices, and</li>
 * <li>generated by a finite number of such elements.</li>
 * </ul> 
 * <p>
 * This class may also be used if the generators are not available, to simply store
 * and manipulate a list of matrices.  Operations requiring generators will in this
 * case simply be skipped over.
 * <p>
 * The discrete group may be a finite or infinite group.  In the latter case,
 * it may have an optional instance of {@link DiscreteGroupConstraint} which is used
 * to determine how to enumerate a finite list of group elements from the generators.
 * <p>
 * Furthermore, the group may have an instance of {@link FiniteStateAutomaton} which
 * complements the role of the constraint by accelerating the solution of the word
 * problem, that is, figuring out whether a particular word in the generators is equivalent
 * to a word which has already been processed.
 * <p>
 * Group elements, including the generators, are represented by instances of
 * {@link DiscreteGroupElement}.
 * <p>
 * There is an optional matrix <i>changeOfBasis</i> which is intended to produce the
 * conjugated group, but it hasn't been implemented systematically.
 * <p>
 * You can also use this class to provide a standard fundamental domain {@link #getDefaultFundamentalRegion()}
 * for the group, an instance of {@link Geometry}.
 * <p>
 * This class also includes
 * a <i>center point</i> for the group.  This is used in various calculations, 
 * such as dirichlet domain calculations, see {@link DirichletDomain}.
 * * <p>
 * The primary method of interest which is not a getter or setter is {@link #update()}.
 * This will calculate the element list anew from the constraint, so should be called
 * whenever the constraint or the generators change.  You may also prefer to use the
 * static method {@link DiscreteGroupUtility#generateElements(DiscreteGroup, DiscreteGroupConstraint)}
 * by hand and keep track of the element list on your own.
 * <p>
 * For examples see the classes in <code>de.jtem.discretegroup.tutorial</code>.
 * 
 * @author Charles Gunn
 *
 */
public class DiscreteGroup {
	
	protected String name = "Unknown discrete Group";
	protected int metric = Pn.EUCLIDEAN,
		dimension = 3,
		maxDirDomOrbitSize = 75;	

	protected boolean isFinite = false,
		isFree = false,
		hasGenerators = true;	

	protected DiscreteGroupElement[]  
	    generators,
		elementList,
		masterList;
	protected FiniteStateAutomaton fsa;
	protected Matrix changeOfBasis = new Matrix();
	protected DiscreteGroupConstraint theConstraint;
	protected double[] centerPoint = Rn.copy(null, P3.originP3);
	protected DiscreteGroupColorPicker	colorPicker;
	
	protected transient boolean changed = true;

	// don't serialize these fields until it's clear how to work with jReality xstream classes
	protected  SceneGraphComponent generatorRepresentations;
	protected Geometry	fundamentalRegion;
	
	protected transient static boolean debug = false;
	
	public DiscreteGroup() {
	}

	protected void init()	{
	}
	
	public void calculateGenerators()	{
	}
	
	public void update()	{
		if (generators == null) calculateGenerators();
		generateElements();
	}

	public void generateElements()	{
		if (generators == null) return;
		elementList = DiscreteGroupUtility.generateElements(this, this.getConstraint());
	}
	
	public Geometry getDefaultFundamentalRegion()	{
		if (isChanged()) update();
		return fundamentalRegion;
	}
	
	public void setDefaultFundamentalDomain(Geometry g) {
		fundamentalRegion = g;
	}

	public void setGenerators(DiscreteGroupElement[] transforms) {
		generators = transforms;
		genTable = null;
		setChanged(true);
		if (generators == null) hasGenerators = false;
	}
	
	public DiscreteGroupElement[] getGenerators()	{
		return generators;
	}
	
	public boolean isDebug() {
		return debug;
	}

	public void setDebug(boolean b) {
		debug = b;
	}
	
	public double[] getCenterPoint() {
		return centerPoint;
	}

	public void setCenterPoint(double[] ds) {
		if (ds == null) return;
		centerPoint[3] = 1.0;
		System.arraycopy(ds,0,centerPoint,0,Math.min(4,ds.length));
	}

	public int getMetric() {
		return metric;
	}

	public void setMetric(int i) {
		metric = i;
	}

	public Matrix getChangeOfBasis() {
		return changeOfBasis;
	}

	public void setChangeOfBasis(Matrix transform) {
		if (changeOfBasis == transform) return;
		changeOfBasis = transform;
		setChanged(true);
	}

	public void setChangeOfBasis(double[] m) {
		System.arraycopy(m, 0, changeOfBasis.getArray(), 0, 16);
//		changeOfBasis.setArray(m);
		setChanged(true);
	}

	public DiscreteGroupConstraint getConstraint() {
		return theConstraint;
	}
	
	public void setConstraint(DiscreteGroupConstraint constraint) {
		theConstraint = constraint;
		setChanged(true);
	}
	
	public FiniteStateAutomaton getFsa() {
		return fsa;
	}

	public void setFsa(FiniteStateAutomaton automaton) {
		fsa = automaton;
	}

	public DiscreteGroupColorPicker getColorPicker() {
		return colorPicker;
	}

	public void setColorPicker(DiscreteGroupColorPicker picker) {
		colorPicker = picker;
	}

	public String getName() {
		return name;
	}

	public void setName(String string) {
		name = string;
	}

	public boolean isFinite() {
		return isFinite;
	}

	public void setFinite(boolean b) {
		isFinite = b;
	}

	public int getDimension() {
		return dimension;
	}

	public void setDimension(int i) {
		dimension = i;
	}

	public boolean hasChanged() {
		return isChanged();
	}

	public void setHasChanged(boolean b) {
		setChanged(b);
	}

	public DiscreteGroupElement[] getElementList() {
		if (isChanged()) update();
		return elementList;
	}

	public void setElementList(DiscreteGroupElement[] elementList) {
		this.elementList = elementList;
		hasGenerators = false;
	}

	public SceneGraphComponent getGeneratorRepresentations() {
		return generatorRepresentations;
	}

	public void setGeneratorRepresentations(SceneGraphComponent gr) {
		generatorRepresentations = gr;
	}

	public boolean isFree() {
		return isFree;
	}

	public void setFree(boolean isFree) {
		this.isFree = isFree;
	}

	transient private  WeakHashMap<DiscreteGroupElement, DiscreteGroupElement>  inverseTable;
	transient protected HashMap<String, DiscreteGroupElement> genTable;

	/**
	 * these methods are used mostly for debugging in DiscreteGroupUtility
	 */
	protected void buildGeneratorHashTable()	{
		if (genTable == null) genTable = new HashMap<String, DiscreteGroupElement>();
		if (generators == null) {
			DiscreteGroupUtility.logger.log(Level.WARNING,"No generators!");
			return;
		}
		for (int i = 0; i<generators.length; ++i)	{
			genTable.put(generators[i].getWord(), generators[i]);
		}
	}
	
	public String getGeneratorInverseWord(String w)	{
		if (genTable == null) buildGeneratorHashTable();
		DiscreteGroupElement dgc = genTable.get(w);
		DiscreteGroupElement inv = getGeneratorInverse(dgc);
		return inv.getWord();
	}

	public DiscreteGroupElement getGeneratorInverse(DiscreteGroupElement dge)	{
		if (inverseTable == null)	{
			inverseTable = new WeakHashMap<DiscreteGroupElement, DiscreteGroupElement>();
			for (DiscreteGroupElement gen : generators)	{
				if (inverseTable.containsKey(gen)) continue;
				String w = gen.getWord();
				String iw = FiniteStateAutomatonUtility.invertString(w);
				for (DiscreteGroupElement gen1 : generators)	{
					if (gen1.getWord().equals(iw)) {
						inverseTable.put(gen, gen1);
						inverseTable.put(gen1,gen);
						break;
					}
				}
				if ( !inverseTable.containsKey(gen)) {
					double[] m2 = Rn.times(null, gen.getArray(), gen.getArray());
					if (! Rn.isIdentityMatrix(m2, 10E-8))
						throw new IllegalStateException("No inverse generator for "+gen);
					inverseTable.put(gen, gen);
				}
			}
		}
		return inverseTable.get(dge);
	}

	public boolean isChanged() {
		return changed;
	}

	public void setChanged(boolean changed) {
		this.changed = changed;
	}

	public int getMaxDirDomOrbitSize() {
		return maxDirDomOrbitSize;
	}

	public void setMaxDirDomOrbitSize(int maxDirDomOrbitSize) {
		this.maxDirDomOrbitSize = maxDirDomOrbitSize;
	}

}
	
